STYLE = """
<style type="text/css">
    .main-style {
        font-size: 12px;
        height: max-content;
        font-family: \'Lucida Sans\', \'Lucida Sans Regular\', \'Lucida Grande\', \'Lucida Sans Unicode\', Geneva, Verdana, sans-serif;
    }
    .text-primary {
        color: #324581;
        font-weight: bold;
    }
    .text-green {
        color: green;
        font-weight: bold;
    }
    .text-red {
        color: red;
        font-weight: bold;
    }
    .text-bolder{
        font-weight: bold;
    }
    .text-lighter{
        font-weight: lighter;
    }
    .fs-small {
        font-size: 0.758rem !important;
    }
    .fs-normal {
        font-size: 0.958rem !important;
    }
    .fs-large {
        font-size: 1.058rem !important;
    }
    .fs-xsmall {
        font-size: 0.558rem !important;
    }
    .avatar {
        vertical-align: middle;
        width: 50x;
        height: 50px;
        border-radius: 50%;
        margin: 2px 0px 2px 0px;
    }
    .div-header{
        display: flex;
        flex-direction: row;
        justify-content: space-between;
        align-items: center;
        padding-top: 1px;
        padding-bottom: 1px;
    }
    .margin-between-horizontal{
        margin: 2px 0px 2px 0px;
    }
    .mt-1{
        margin-top: 1px;
    }
    .title-table-category  {
        font-size: 10px;
        font-weight: bold;
    }
    .title-table-resume {
        font-size: 10px;
        font-weight: bold;
        text-align: center;
    }

    .div-table {
        display: flex;
        flex-direction: column;
    }

    .table {
        width: 100%;
        height: 100%;
        border-collapse: collapse;
        text-align: center;
    }

    .tr-table-top {
        border-bottom: 0.5px solid gray;
        border-top: 0.5px solid gray;
    }

    .tr-table-center {
        border-top: 0.0px solid gray;
        border-bottom: 0.0px solid gray;
    }

    .tr-table-bottom {
        border-top: 0px solid gray;
    }

    .tr-table-border {
        border-top: 1px dotted gray;
        border-bottom: 1px dotted gray;
    }

    .sub-header-info {
        display: flex;
        flex-direction: column;
        justify-content: end;
        width: 100%;
    }

    .sub-header-info-label {
        width: 20%;
        text-align: start;
    }

    .div-sub-header {
        border-top: 2px solid gray;
    }

    .sub-header-text {
        display: flex;
        flex-direction: row;
        align-items: center;
        justify-content: start;
    }

    .text-center {
        text-align: center;
    }

    .text-left {
        text-align: left;
    }

    .text-right {
        text-align: right;
    }

    .footer-total {
        display: flex;
        justify-content: end;
        align-items: center;
        width: 100%;
        margin-top: 5px;
    }

    .ml-4 {
        margin-left: 2rem;
    }

    .flex-column {
        display: flex;
        flex-direction: column;
    }

    .flex {
        display: flex;
    }

    .td-table-left {
        text-align: left;
        font-size: 10px;
        padding-left: 10px;
        border: 0.5px solid gray;
    }

    .td-table-right {
        text-align: right;
        font-size: 10px;
        padding-right: 10px;
        border: 0.5px solid gray;
    }
</style>
"""  # noqa


from sqlalchemy import orm, UniqueConstraint

from viggocore.database import db
from viggocore.common.subsystem import entity


class PolicyModule(entity.Entity, db.Model):

    attributes = ['capability_module_id', 'role_id']
    attributes += entity.Entity.attributes

    capability_module_id = db.Column(
        db.CHAR(32), db.ForeignKey("capability_module.id"),
        nullable=False)
    capability_module = orm.relationship(
        'CapabilityModule',
        backref=orm.backref('policy_module_capability_module'))
    role_id = db.Column(db.CHAR(32), db.ForeignKey("role.id"), nullable=False)
    role = orm.relationship(
        'Role', backref=orm.backref('policy_module_role'))

    __table_args__ = (
        UniqueConstraint(
            'capability_module_id', 'role_id',
            name='pm_capability_module_id_role_id_uk'),)

    def __init__(self, id, capability_module_id, role_id,
                 active=True, created_at=None, created_by=None,
                 updated_at=None, updated_by=None, tag=None):
        super().__init__(id, active, created_at, created_by,
                         updated_at, updated_by, tag)
        self.capability_module_id = capability_module_id
        self.role_id = role_id

    @classmethod
    def individual(cls):
        return 'policy_module'

    @classmethod
    def collection(cls):
        return 'policy_modules'



import json

import flask

from viggocore.common import exception, utils
from viggocore.common.exception import BadRequest
from viggocore.common.subsystem import controller
from viggocore.subsystem.domain.functions.controller import (
    consulta_cpf_cnpj)
from viggocore.subsystem.image.resource import QualityImage
from datetime import datetime
import datetime as datetime2
from viggocore.common.subsystem import entity


class Controller(controller.Controller):

    def __init__(self, manager, resource_wrap, collection_wrap):
        super(Controller, self).__init__(
            manager, resource_wrap, collection_wrap)

    def _get_name_in_args(self):
        name = flask.request.args.get('name', None)
        if not name:
            raise BadRequest()
        return name

    def domain_by_name(self):
        try:
            name = self._get_name_in_args()
            domain = self.manager.domain_by_name(domain_name=name)
            domain_dict = domain.to_dict()
            if 'settings' in domain_dict.keys():
                domain_dict.pop('settings')
            response = {self.resource_wrap: domain_dict}
            return flask.Response(response=json.dumps(response, default=str),
                                  status=200,
                                  mimetype="application/json")
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

    def domain_logo_by_name(self):
        try:
            kwargs = {}
            quality = flask.request.args.get('quality', None)
            kwargs['quality'] = \
                QualityImage[quality] if quality else QualityImage.med
            name = self._get_name_in_args()
            folder, filename = self.manager.domain_logo_by_name(
                domain_name=name, **kwargs)
            return flask.send_from_directory(folder, filename)
        except KeyError:
            return self.get_bad_request('Unknown Quality')
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

    def get_token_id(self):
        return flask.request.headers.get('token')

    def get_token(self, token_id):
        return self.manager.api.tokens().get(id=token_id)

    def get_domain(self, domain_id):
        return self.manager.api.domains().get(id=domain_id)

    def get_domain_id_from_token(self, token):
        user = self.manager.api.users().get(id=token.user_id)
        return user.domain_id

    def get_domain_id(self):
        token = self.get_token(self.get_token_id())
        domain_id = self.get_domain_id_from_token(token)
        return domain_id

    def upload_logo(self, id):
        try:
            token = flask.request.headers.get('token')
            file = flask.request.files.get('file', None)
            if not file:
                raise exception.BadRequest('ERROR! File not found in request.')

            image = self.manager.upload_logo(id=id, token=token, file=file)

            response = {'image': image.to_dict()}
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

        return flask.Response(response=utils.to_json(response),
                              status=201,
                              mimetype="application/json")

    def remove_logo(self, id):
        try:
            self.manager.remove_logo(id=id)
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

        return flask.Response(response=None,
                              status=204,
                              mimetype="application/json")

    def register(self):
        try:
            data = flask.request.get_json()

            username = data.get('username', 'admin').lower()
            email = data.get('email', None)
            password = data.get('password', None)
            domain_name = data.get('domain_name', None)
            domain_display_name = data.get('domain_display_name', None)
            application_name = data.get('application_name', None)

            self.manager.register(
                username=username, email=email,
                password=password, domain_name=domain_name,
                domain_display_name=domain_display_name,
                application_name=application_name)
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

        return flask.Response(response=None,
                              status=204,
                              mimetype="application/json")

    def activate(self, id1, id2):
        try:
            token_id = flask.request.headers.get('token')
            domain = self.manager.activate(
                token_id=token_id, domain_id=id1, user_admin_id=id2)

            response = {'domain': domain.to_dict()}
            return flask.Response(response=utils.to_json(response),
                                  status=200,
                                  mimetype="application/json")
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

    def update_settings(self, id):
        try:
            data = flask.request.get_json()

            settings = self.manager.update_settings(id=id, **data)
            response = {'settings': settings}

            return flask.Response(response=utils.to_json(response),
                                  status=200,
                                  mimetype="application/json")
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

    def _get_keys_from_args(self):
        keys = flask.request.args.get('keys')
        if not keys:
            raise exception.BadRequest(
                'ERRO! The keys parameter was not passed correctly')
        return list(filter(None, keys.split(',')))

    def remove_settings(self, id):
        try:
            keys = self._get_keys_from_args()
            kwargs = {'keys': keys}

            settings = self.manager.remove_settings(id=id, **kwargs)
            response = {'settings': settings}

            return flask.Response(response=utils.to_json(response),
                                  status=200,
                                  mimetype="application/json")
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

    def get_domain_settings_by_keys(self, id):
        try:
            keys = self._get_keys_from_args()
            kwargs = {'keys': keys}

            settings = self.manager.get_domain_settings_by_keys(
                id=id, **kwargs)
            response = {'id': id, 'settings': settings}

            return flask.Response(response=utils.to_json(response),
                                  status=200,
                                  mimetype="application/json")
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

    def send_email_activate_account(self):
        try:
            data = flask.request.get_json()

            self.manager.send_email_activate_account(**data)
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

        return flask.Response(response=None,
                              status=204,
                              mimetype="application/json")

    def get_domain_settings_by_publicas(self, id):
        try:
            kwargs = {'keys': ['public']}

            settings = self.manager.get_domain_settings_by_keys(
                id=id, **kwargs)
            response = {'id': id, 'settings': settings}

            return flask.Response(response=utils.to_json(response),
                                  status=200,
                                  mimetype="application/json")
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

    def get_files_size(self, id):
        try:
            files_size = self.manager.get_files_size(id=id)
            response = {'id': id, 'sizes': files_size}

            return flask.Response(response=utils.to_json(response),
                                  status=200,
                                  mimetype="application/json")
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

    # return start and end
    def _get_page_and_page_size(self, page, page_size):
        if page < 0:
            raise exception.BadRequest(
                'the page parameter must be greater than -1')
        if page_size < 1:
            raise exception.BadRequest(
                'the page_size parameter must be greater than 0')

        if page == 0:
            return (page, page_size)
        else:
            start = page * page_size
            end = start + page_size
            return (start, end)

    def _convert_de_ate(self, **kwargs):
        de = kwargs.pop('de', '2020-01-01-0300')
        ate = kwargs.pop('ate', '9020-01-01-0300')

        if de and ate:
            try:
                inicio = datetime.strptime(de.replace(' ', '+'), '%Y-%m-%d%z')
                fim = datetime.strptime(ate.replace(' ', '+'), '%Y-%m-%d%z') +\
                    datetime2.timedelta(days=1)

            except Exception:
                inicio = datetime.strptime(de, entity.DATE_FMT)
                fim = datetime.strptime(ate, entity.DATE_FMT) +\
                    datetime2.timedelta(days=1)
            kwargs['de'] = inicio
            kwargs['ate'] = fim

        return kwargs

    def get_usage_info_by_domain(self):
        filters = self._filters_parse()
        filters = self._filters_cleanup(filters)

        try:
            filters = self._parse_list_options(filters)
            filters = self._convert_de_ate(**filters)

            require_pagination = filters.pop('require_pagination', False)
            page = filters.pop('page', None)
            page_size = filters.pop('page_size', None)
            domain_name = filters.pop('domain_name', None)
            active = filters.pop('active', None)

            entities, totals = self.manager.get_usage_info_by_domain(**filters)

        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)
        except ValueError:
            raise exception.BadRequest('page or page_size is invalid')

        if domain_name is not None:
            if '%' in domain_name:
                domain_name = domain_name.replace('%', '')
                entities = list(
                    filter(lambda x: domain_name in x['domain_name'],
                           entities))
            else:
                entities = list(
                    filter(lambda x: domain_name == x['domain_name'],
                           entities))

        if active is not None:
            entities = list(
                filter(lambda x: x['active'] == active, entities))

        entities = sorted(entities, key=lambda item: item['domain_name'])

        total = len(entities)

        if require_pagination and None not in [page, page_size]:
            page = int(page)
            page_size = int(page_size)

            start, end = self._get_page_and_page_size(page, page_size)

            response = {self.collection_wrap: entities[start:end],
                        'totals': totals}

            response.update({'pagination': {'page': page,
                                            'page_size': page_size,
                                            'total': total}})
        else:
            response = {self.collection_wrap: entities, 'totals': totals}

        return flask.Response(response=utils.to_json(response),
                              status=200,
                              mimetype="application/json")

    def get_roles(self, id: str):
        try:
            roles = self.manager.get_roles(id=id)
            response = {"roles": [role.to_dict() for role in roles]}

        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

        return flask.Response(response=utils.to_json(response),
                              status=200,
                              mimetype="application/json")

    def check_permission(self, id: str):
        filters = self._filters_parse()
        filters = self._filters_cleanup(filters)

        try:
            filters = self._parse_list_options(filters)
            url = filters.get('url', None)
            method = filters.get('method', None)
            if not url or not method:
                raise exception.BadRequest('url and method are required')

            has_permission, route = self.manager.check_permission(
                id=id, url=url, method=method)
            response = {
                "domain_id": id,
                "route": route.to_dict(),
                "has_permission": has_permission}

        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

        return flask.Response(response=utils.to_json(response),
                              status=200,
                              mimetype="application/json")

    def list(self):
        filters = self._filters_parse()
        # filters = self._filters_cleanup(filters)

        try:
            filters = self._parse_list_options(filters)
            entities = self.manager.list(**filters)

            with_pagination = False
            require_pagination = filters.get('require_pagination', False)
            page = filters.get('page', None)
            page_size = filters.get('page_size', None)

            if (page and page_size is not None) and require_pagination is True:
                with_pagination = True
                count = self.manager.count(**(self._clean_filters(**filters)))
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)
        except ValueError:
            raise exception.BadRequest('page or page_size is invalid')

        collection = self._entities_to_dict(
            entities, self._get_include_dicts(filters))

        response = {self.collection_wrap: collection}

        if with_pagination:
            response.update({'pagination': {'page': int(page),
                                            'page_size': int(page_size),
                                            'total': count}})

        return flask.Response(response=utils.to_json(response),
                              status=200,
                              mimetype="application/json")

    def consulta_cpf_cnpj(self):
        return consulta_cpf_cnpj.consulta_cpf_cnpj(self)



from apscheduler.schedulers.background import BackgroundScheduler


class Scheduler(BackgroundScheduler):

    def __init__(self):
        super().__init__()
        self.start()

    def schedule(self, callback, **kwargs):
        self.add_job(callback, 'cron', **kwargs)

    def hourly(self, callback, minute=0):
        self.add_job(callback, 'cron', minute=minute)

    def daily(self, callback, hour=0, minute=0):
        self.add_job(callback, 'cron', hour=hour, minute=minute)

    def weekly(self, callback, day='mon', hour=0, minute=0):
        if day not in ('mon', 'tue', 'wed', 'thu', 'fri', 'sat', 'sun'):
            raise ValueError
        self.add_job(
            callback, 'cron', day_of_week=day, hour=hour, minute=minute)

    def monthly(self, callback, day=1, hour=0, minute=0):
        self.add_job(callback, 'cron', day=day, hour=hour, minute=minute)



import flask
import uuid

from viggocore.common import exception
from viggocore.common.subsystem.apihandler import Api, ApiHandler


class Request(flask.Request):

    # TODO(samueldmq): find a better place to put this utility method
    def _check_uuid4(self, uuid_str):
        if len(uuid_str) != 32:
            return False
        try:
            return uuid.UUID(uuid_str, version=4)
        except ValueError:
            return False

    @property
    def url(self):
        path_info = flask.request.environ['PATH_INFO'].rstrip('/')
        path_bits = [
            '<id>' if self._check_uuid4(i) else i for i in path_info.split('/')
        ]

        if path_bits.count('<id>') > 1:
            pos = 0
            qty_id = 1
            for bit in path_bits:
                if bit == '<id>':
                    path_bits[pos] = '<id' + str(qty_id) + '>'
                    qty_id += 1
                pos += 1
        return '/'.join(path_bits)

    @property
    def token(self):
        return flask.request.headers.get('token')


class RequestManager(object):

    def __init__(self, api_handler: ApiHandler):
        self.api_handler = api_handler

    def before_request(self):
        api: Api = self.api_handler.api()
        if flask.request.method == 'OPTIONS':
            return

        # Short-circuit if accessing the root URL,
        # which will just return the version
        # TODO(samueldmq): Do we need to create a subsystem just for this ?
        if not flask.request.url:
            return

        routes = api.routes().list(url=flask.request.url,
                                   method=flask.request.method)
        if not routes:
            msg = 'Route not found'
            return flask.Response(response=msg, status=404)
        route = routes[0]

        if not route.active:
            msg = 'Route is inactive'
            return flask.Response(response=msg, status=410)

        if route.bypass:
            return

        token_id = flask.request.token

        if not token_id:
            msg = 'Token is required'
            return flask.Response(response=msg, status=401)

        try:
            token = api.tokens().get(id=token_id)
        except exception.NotFound:
            msg = 'Token not found'
            return flask.Response(response=msg, status=401)

        can_access = api.users().authorize(user_id=token.user_id, route=route)

        if not can_access:
            msg = 'You do not have permission'
            return flask.Response(response=msg, status=403)
        return



from sqlalchemy import orm

from viggocore.common.subsystem import entity
from viggocore.database import db


class LogSystem(entity.Entity, db.Model):

    attributes = ['domain_id', 'user_id', 'user_name', 'user_password',
                  'justificativa', 'entidade_ids', 'entidade_nome',
                  'json_anterior', 'json_posterior']
    attributes += entity.Entity.attributes

    # ligações
    domain_id = db.Column(
        db.CHAR(32), db.ForeignKey('domain.id'), nullable=False)
    user_id = db.Column(
        db.CHAR(32), db.ForeignKey('user.id'), nullable=False)

    # colunas
    user_name = db.Column(db.String(80), nullable=True)
    user_password = db.Column(db.String(64), nullable=True)
    justificativa = db.Column(db.String(500), nullable=True)
    entidade_ids = db.Column(db.Text, nullable=False)
    entidade_nome = db.Column(db.String(60), nullable=False)
    json_anterior = db.Column(db.Text, nullable=False)
    json_posterior = db.Column(db.Text, nullable=False)

    # backrefs
    user = orm.relationship(
        "User", backref=orm.backref('log_system_user'),
        viewonly=True)

    def __init__(self, id, domain_id, user_id, user_name, entidade_ids,
                 entidade_nome, json_anterior, json_posterior,
                 user_password=None, justificativa=None,
                 active=True, created_at=None, created_by=None,
                 updated_at=None, updated_by=None, tag=None):
        super().__init__(id, active, created_at, created_by, updated_at,
                         updated_by, tag)
        self.domain_id = domain_id
        self.user_id = user_id
        self.user_name = user_name
        self.entidade_ids = entidade_ids
        self.entidade_nome = entidade_nome
        self.json_anterior = json_anterior
        self.json_posterior = json_posterior
        self.user_password = user_password
        self.justificativa = justificativa

    @classmethod
    def individual(self):
        return 'log_system'

    @classmethod
    def collection(cls):
        return 'log_systems'



from sqlalchemy import orm, UniqueConstraint

from viggocore.database import db
from viggocore.common.subsystem import entity


class CapabilityModule(entity.Entity, db.Model):

    attributes = ['module_id', 'application_id']
    attributes += entity.Entity.attributes

    module_id = db.Column(
        db.CHAR(32), db.ForeignKey("module.id"), nullable=False)
    module = orm.relationship(
        'Module', backref=orm.backref('capability_module_module'))
    application_id = db.Column(
        db.CHAR(32), db.ForeignKey("application.id"), nullable=False)
    application = orm.relationship('Application', backref=orm.backref(
        'capability_module_application'))

    __table_args__ = (
        UniqueConstraint(
            'module_id', 'application_id',
            name='cp_module_id_application_id_uk'),)

    def __init__(self, id, module_id, application_id,
                 active=True, created_at=None, created_by=None,
                 updated_at=None, updated_by=None, tag=None):
        super().__init__(id, active, created_at, created_by,
                         updated_at, updated_by, tag)
        self.module_id = module_id
        self.application_id = application_id

    @classmethod
    def individual(cls):
        return 'capability_module'

    @classmethod
    def collection(cls):
        return 'capability_modules'



from sqlalchemy import orm

from viggocore.database import db
from viggocore.common.subsystem import entity


class ProjectCost(entity.Entity, db.Model):

    attributes = ['name', 'cost']
    attributes += entity.Entity.attributes

    name = db.Column(db.String(30), nullable=False)
    cost = db.Column(db.Numeric(13, 3), nullable=False)

    def __init__(self, id, name, cost,
                 active=True, created_at=None, created_by=None,
                 updated_at=None, updated_by=None, tag=None):
        super().__init__(id, active, created_at, created_by,
                         updated_at, updated_by, tag)
        self.name = name
        self.cost = cost

    @classmethod
    def individual(cls):
        return 'project_cost'



import flask
import json

from viggocore.common import utils
from viggocore.common.subsystem import controller
from viggocore.common import exception
from viggocore.subsystem.user.email import TypeEmail


class Controller(controller.Controller):

    def __init__(self, manager, resource_wrap, collection_wrap):
        super(Controller, self).__init__(
            manager, resource_wrap, collection_wrap)

    def get_token_id(self):
        return flask.request.headers.get('token')

    def get_token(self, token_id):
        return self.manager.api.tokens().get(id=token_id)

    def get_domain(self, domain_id):
        return self.manager.api.domains().get(id=domain_id)

    def get_domain_id_from_token(self, token):
        user = self.manager.api.users().get(id=token.user_id)
        return user.domain_id

    def get_domain_id(self):
        token = self.get_token(self.get_token_id())
        domain_id = self.get_domain_id_from_token(token)
        return domain_id

    def restore(self):
        if not flask.request.is_json:
            return flask.Response(
                response=exception.BadRequestContentType.message,
                status=exception.BadRequestContentType.status)

        data = flask.request.get_json()

        try:
            self.manager.restore(**data)
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

        return flask.Response(response=None,
                              status=200,
                              mimetype="application/json")

    def reset_password(self, id):
        if not flask.request.is_json:
            return flask.Response(
                response=exception.BadRequestContentType.message,
                status=exception.BadRequestContentType.status)

        data = flask.request.get_json()
        try:
            user = self.manager.reset(id=id, **data)
            self.manager.api.tokens().deletar_tokens(**{
                'user_id': id,
                'domain_id': user.domain_id})
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

        return flask.Response(response=None,
                              status=204,
                              mimetype="application/json")

    def reset_my_password(self):
        if not flask.request.is_json:
            return flask.Response(
                response=exception.BadRequestContentType.message,
                status=exception.BadRequestContentType.status)

        data = flask.request.get_json()
        try:

            token = self.get_token(self.get_token_id())
            if not token:
                raise exception.BadRequest()

            self.manager.reset(id=token.user_id, **data)
            self.manager.api.tokens().delete(id=token.id)
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

        return flask.Response(response=None,
                              status=204,
                              mimetype="application/json")

    def routes(self):
        if not flask.request.is_json:
            return flask.Response(
                response=exception.BadRequestContentType.message,
                status=exception.BadRequestContentType.status)

        token = self.manager.api.tokens().get(
            id=flask.request.headers.get('token'))
        filters = self._filters_parse()
        try:
            routes = self.manager.routes(user_id=token.user_id, **filters)
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

        response = {"routes": [route.to_dict() for route in routes]}

        return flask.Response(response=json.dumps(response, default=str),
                              status=200,
                              mimetype="application/json")

    def upload_photo(self, id, **kwargs):
        try:
            file = flask.request.files.get('file', None)
            if not file:
                raise exception.BadRequest()

            token = self.get_token(self.get_token_id())
            domain_id = self.get_domain_id_from_token(token)
            user_id = token.user_id

            if not (domain_id and user_id):
                raise exception.BadRequest()

            kwargs['domain_id'] = domain_id
            kwargs['user_id'] = user_id
            kwargs['type_image'] = 'UserPhoto'
            image = self.manager.api.images().create(file=file, **kwargs)

            kwargs.pop('type_image')
            kwargs['photo_id'] = image.id
            self.manager.upload_photo(id=id, **kwargs)

            response = {'image': image.to_dict()}
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

        return flask.Response(response=utils.to_json(response),
                              status=201,
                              mimetype="application/json")

    def delete_photo(self, id):
        try:
            self.manager.delete_photo(id=id)
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

        return flask.Response(response=None,
                              status=204,
                              mimetype="application/json")

    def update_password(self, id):
        data = flask.request.get_json()
        try:
            token = self.get_token(self.get_token_id())
            password = data.get('password', None)
            old_password = data.get('old_password', None)

            if token.user_id != id:
                error = exception.ViggoCoreException()
                error.status = 401
                error.message = "Not Authorized"
                raise error
            if not password or not old_password:
                raise exception.BadRequest()
            self.manager.update_password(
                id=id, password=password, old_password=old_password)
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

        return flask.Response(response=None,
                              status=204,
                              mimetype="application/json")

    def notify(self, id):
        data = flask.request.get_json()
        try:
            type_email = TypeEmail.value_of(data.get('type_email', None))
            token = self.get_token(self.get_token_id())

            if not type_email or not token:
                raise exception.BadRequest()

            self.manager.notify(id=id, type_email=type_email, token=token)
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

        return flask.Response(response=None,
                              status=204,
                              mimetype="application/json")

    def roles(self, id):
        try:
            roles = self.manager.roles(user_id=id)
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

        response = {"roles": [role.to_dict() for role in roles]}

        return flask.Response(response=json.dumps(response, default=str),
                              status=200,
                              mimetype="application/json")

    def update_settings(self, id):
        try:
            data = flask.request.get_json()

            settings = self.manager.update_settings(id=id, **data)
            response = {'settings': settings}

            return flask.Response(response=utils.to_json(response),
                                  status=200,
                                  mimetype="application/json")
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

    def _get_keys_from_args(self):
        keys = flask.request.args.get('keys')
        if not keys:
            raise exception.BadRequest(
                'ERRO! The keys parameter was not passed correctly')
        return list(filter(None, keys.split(',')))

    def remove_settings(self, id):
        try:
            keys = self._get_keys_from_args()
            kwargs = {'keys': keys}

            settings = self.manager.remove_settings(id=id, **kwargs)
            response = {'settings': settings}

            return flask.Response(response=utils.to_json(response),
                                  status=200,
                                  mimetype="application/json")
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

    def get_user_settings_by_keys(self, id):
        try:
            keys = self._get_keys_from_args()
            kwargs = {'keys': keys}

            settings = self.manager.get_user_settings_by_keys(
                id=id, **kwargs)
            response = {'id': id, 'settings': settings}

            return flask.Response(response=utils.to_json(response),
                                  status=200,
                                  mimetype="application/json")
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)



from typing import List

from viggocore.common import exception, utils
from viggocore.common.input import RouteResource
from viggocore.common.subsystem import manager, operation
from viggocore.subsystem.capability.resource import Capability
from viggocore.subsystem.route.resource import Route


class CreateCapabilities(operation.Operation):
    """ Receive a list of resources for create capabilities
        resource -> ('url','method') """

    def _routes_without_capabilities(self, capabilities: List[Capability],
                                     routes: List[Route]) -> List[Route]:
        filtered_routes = []
        routes_capabilities_ids = [c.route_id for c in capabilities]
        for route in routes:
            if route.id not in routes_capabilities_ids:
                filtered_routes.append(route)

        return filtered_routes

    def _filter_routes(self, resources: List[RouteResource],
                       routes: List[Route]) -> List[str]:
        routes_ids = [route.id for route in routes
                      if (route.url, route.method) in resources]
        return routes_ids

    def _create_capability(self, application_id: str,
                           route_id: str) -> Capability:
        return Capability(utils.random_uuid(),
                          route_id=route_id,
                          application_id=application_id)

    def pre(self, session, id: str, **kwargs) -> bool:
        self.application_id = id
        resources = kwargs.get('resources', None)
        if not resources:
            raise exception.OperationBadRequest()

        routes = self.manager.api.routes().list(sysadmin=False, active=True)
        capabilities = self.manager.api.capabilities().list(
            application_id=self.application_id)
        routes_without_capabilities = \
            self._routes_without_capabilities(capabilities, routes)

        self.routes_ids = self._filter_routes(resources,
                                              routes_without_capabilities)

        return True
        # return self.driver.get(id, session=session) is not None

    def do(self, session, **kwargs) -> None:
        capabilities = [self._create_capability(self.application_id, route_id)
                        for route_id in self.routes_ids]

        session.bulk_save_objects(capabilities)


class Delete(operation.Delete):

    def pre(self, session, id, **kwargs):
        super().pre(session, id=id)
        policies = self.manager.api.policies().list(capability_id=id)
        if policies:
            message = 'You can\'t remove this capability because' + \
                ' there are policies associated'
            raise exception.BadRequest(message)
        return True


class Manager(manager.Manager):

    def __init__(self, driver):
        super().__init__(driver)
        self.create_capabilities = CreateCapabilities(self)
        self.delete = Delete(self)



import enum


class STYLE_PAGINATION_TYPE(enum.Enum):
    RETRATO = """
    <style>
        @page {
            size: A4;
            margin: 10mm;
        }
        @page {
            @bottom-right {
                content: "Página " counter(page) " de " counter(pages);
                font-size: 12px;
            }
        }
    </style>"""  # noqa

    PAISAGEM = """
    <style>
        @page {
            size: A4 landscape;
            margin: 10mm;
        }
        @page {
            @bottom-right {
                content: "Página " counter(page) " de " counter(pages);
                font-size: 12px;
            }
        }
    </style>"""  # noqa


from enum import Enum
from datetime import datetime
from viggocore.common.subsystem import operation
from viggocore.common import exception, utils as core_utils
from viggocore.common.subsystem.gerar_pdf import (
    relatorio_utils as ru,
    layout_padrao_html as layout)


class ExportPdf(operation.Operation):

    def _treat_data_type(self, data, col):
        data_type = col.get('type', 'STRING')
        result = ''

        if data_type == 'STRING':
            result = data if data is not None else '-'
            if data is not None:
                result = data.name if isinstance(data, Enum) else data
        elif data_type == 'BOOLEAN':
            result = 'Sim' if data is True else 'Não'
        elif data_type == 'DATE':
            result = datetime.strftime(data, '%d/%m/%Y') \
                if data is not None \
                else '-'
        elif data_type == 'DATETIME':
            result = datetime.strftime(data, '%d/%m/%Y %H:%M') \
                if data is not None \
                else '-'
        elif data_type == 'CURRENCY' or data_type == 'FLOAT':
            decimals = col.get('decimals', None)
            if decimals is not None:
                result = core_utils.to_decimal_n(data, decimals) \
                    if data is not None \
                    else core_utils.to_decimal_n('0', decimals)
            else:
                result = core_utils.to_decimal(data) \
                    if data is not None \
                    else core_utils.to_decimal('0')
        elif data_type == 'INT':
            result = int(data) if data is not None else 0

        return result

    def _montar_tr(self, columns, entity):
        tr_html = ''
        td_html = ''
        for col in columns:
            value = ''
            fields = col.get('field', None)
            if fields is not None:
                fields = fields.split('.')
                if len(fields) > 1:
                    value = None if entity is None else entity
                    if entity is not None:
                        for field in fields:
                            if value is not None:
                                value = value.__getattribute__(field)
                else:
                    value = entity.__getattribute__(fields[0])

                value = self._treat_data_type(value, col)
            td_html += layout.TD.format(
                width=col.get('width', '1%'),
                value=value)
        if len(td_html) > 0:
            tr_html = layout.TR.format(tds=td_html)

        return tr_html

    def _montar_tr_title(self, columns):
        td_html = ''
        for col in columns:
            td_html += layout.TD_BOLD.format(
                width=col.get('width', '1%'),
                value=col.get('header_name', ''))
        return layout.TR.format(tds=td_html)

    # TODO(JorgeSilva): verificar a necessidade de mapear os filtros aplicados
    def _get_filtros(self, **kwargs):
        return ''

    def _get_domain_display_name(self, entities):
        domain_display_name = ''
        info_empresa = ''
        try:
            domain = entities[0].domain
        except Exception:
            domain = self.manager.api.domains().get(id=entities[0].domain_id)

        # verifica se o display_name do domínio não está vazio
        if not core_utils.is_empty_or_blank(domain.display_name):
            domain_display_name = domain.display_name

        if self.com_endereco:
            address = ''
            address2 = ''
            if len(domain.addresses) > 0:
                addr = domain.addresses[0]
                address = (
                    f'{addr.logradouro}, {addr.numero}, {addr.bairro}, ' +
                    f'{addr.municipio.nome}/{addr.municipio.sigla_uf}')
                address2 = f'CEP. {ru.formatar_cep(addr.cep)}'
                if domain.doc is not None:
                    address2 = f'{address2} - CNPJ: ' \
                        if len(domain.doc) > 11 \
                        else f'{address2} - CPF: '
                    address2 = address2 + \
                        ru.apply_mask_cpf_cnpj(domain.doc)

            if address != '':
                info_empresa = "<br>".join([
                    address, address2, ''])

        return (domain_display_name, info_empresa)

    def _gerar_html(self, entities, columns):
        body = layout.BODY
        data_solicitacao = self.data_solicitacao.strftime('%d/%m/%Y, %H:%M')
        domain_display_name, info_empresa = self._get_domain_display_name(
            entities)

        # montar o html do título de cada coluna da tabela
        tr_title = self._montar_tr_title(columns)

        # montar o html das linhas da tabela
        trs = ''
        for entity in entities:
            trs += self._montar_tr(columns, entity)

        # montar o footer do html
        footer = layout.TR_FOOTER \
            .format(count=f'Total de linhas: {len(entities)}')

        img_url = ru.get_logo_path(self.manager, self.domain_id)

        # monta o header
        header = ru.montar_header(
            img_url=img_url,
            nome_empresa=domain_display_name,
            info_empresa=info_empresa,
            titulo=self.title,
            data_emissao=data_solicitacao,
            filtros=self._get_filtros(),
            domain_name_style=self.domain_name_style,
            domain_info_style=self.domain_info_style,
            title_style=self.title_style)

        # monta o body (este é específico para cada relatório)
        body = body.format(tr_title=tr_title, trs=trs)

        # joga o header e o body formatados dentro do template
        template = ru.montar_template(
            header=header, body=body, op_style=7, footer=footer,
            style_pagination=self.style_pagination)

        return template

    def pre(self, **kwargs):
        # pega a flag com_endereco responsável por contatenar o nome
        # do domínio com o seu endereço caso a flag seja passada
        self.com_endereco = kwargs.pop('com_endereco', False)
        # pega a data de solicitação de relatório para colocar no pdf
        self.data_solicitacao = ru.get_data_solicitacao_dt(**kwargs)
        # pega o título que deseja que apareça no relatório
        self.title = kwargs.get('title', 'Exportação em PDF')
        # pega o domain_id para ter as informações do domínio e conseguir
        # puxar a logo para montar o cabeçalho do PDF
        self.domain_id = kwargs.get('domain_id', None)
        # verifica se o domain_id foi passado na requisição ou não
        if self.domain_id is None:
            raise exception.BadRequest('É obrigatório passar o domain_id.')

        # caso seja passado um style_pagination então o backend irá
        # aplicar ele, caso não seja passado então será aplicado o
        # RETRATO como padrão
        style_pagination = kwargs.get('style_pagination', None)
        if style_pagination is None:
            self.style_pagination = ru.STYLE_PAGINATION_TYPE.RETRATO
        else:
            self.style_pagination = ru.STYLE_PAGINATION_TYPE[style_pagination]

        # pega as variáveis do style de cada parte do meio do header.
        # Se não for passado vai manter o padrão
        self.domain_name_style = kwargs.get('domain_name_style', None)
        self.domain_info_style = kwargs.get('domain_info_style', None)
        self.title_style = kwargs.get('title_style', None)

        return True

    def do(self, session, **kwargs):
        filters = kwargs.get('filters', {})
        columns = kwargs.get('columns', [])

        self.filtros = self._get_filtros(**filters)

        entities = self.manager.list(session=session, **filters)
        if type(entities) is tuple:
            entities = entities[0]

        if len(entities) > 0:
            html = self._gerar_html(entities, columns).replace('\n', '')
        else:
            raise exception.BadRequest(
                'Não existe nada para essa filtragem.')

        # chama a função que gera o pdf e trata a resposta
        response = ru.gerar_pdf(html)
        return response



from sqlalchemy import UniqueConstraint

from viggocore.database import db
from viggocore.common.subsystem import entity


class DomainAndTableSequence(entity.Entity, db.Model):

    attributes = ['domain_id', 'table_id', 'name', 'value']
    attributes += entity.Entity.attributes

    domain_id = db.Column(
        db.CHAR(32), db.ForeignKey('domain.id'), nullable=False)
    table_id = db.Column(db.String(100), nullable=False)
    name = db.Column(db.String(100), nullable=False)
    value = db.Column(db.Numeric(10), default=0, nullable=False)

    __table_args__ = (
        UniqueConstraint(
            'name',
            'table_id',
            'domain_id',
            name='domain_and_table_sequence_name_table_id_domain_id_uk'),)

    def __init__(self, id, domain_id, table_id, name, value,
                 active=True, created_at=None, created_by=None,
                 updated_at=None, updated_by=None, tag=None):
        super().__init__(id, active, created_at, created_by,
                         updated_at, updated_by, tag)
        self.domain_id = domain_id
        self.table_id = table_id
        self.name = name
        self.value = value

    @classmethod
    def individual(cls):
        return 'domain_and_table_sequence'

    def is_stable(self):
        domain_id_stable = self.domain_id is not None
        table_id_stable = self.table_id is not None
        name_stable = self.name is not None
        value_stable = self.value is not None and self.value >= 0

        return domain_id_stable and table_id_stable \
            and name_stable and value_stable



from typing import List, Optional
from viggocore.common import exception, utils
from viggocore.common.subsystem import manager
from viggocore.common.subsystem import operation
from viggocore.subsystem.role.resource import Role, RoleDataViewType
from viggocore.subsystem.policy.resource import Policy


class Create(operation.Create):

    def pre(self, session, **kwargs):
        if kwargs.get('data_view', None) is None:
            kwargs['data_view'] = RoleDataViewType.DOMAIN
        return super().pre(session, **kwargs)


class CreatePolicies(operation.Operation):

    def _filter_capabilities(self, routes, capabilities, policies, resources):
        policies_capabilities_ids = [policy.capability_id
                                     for policy in policies]
        capabilities = [c for c in capabilities
                        if c.id not in policies_capabilities_ids]
        routes_ids = [route.id for route in routes
                      if (route.url, route.method) in resources]

        filtered_capabilities = [capability for capability in capabilities
                                 if capability.route_id in routes_ids]
        return filtered_capabilities

    def _create_policy(self, role_id: str, capability_id: str) -> Policy:
        return Policy(id=utils.random_uuid(),
                      role_id=role_id,
                      capability_id=capability_id)

    def pre(self, session, id: str, **kwargs) -> bool:
        self.role_id = id
        self.application_id = kwargs.get('application_id', None)
        resources = kwargs.get('resources', None)

        if ((not (self.role_id and self.application_id)) or resources is None):
            raise exception.OperationBadRequest()

        try:
            self.manager.api.applications().get(id=self.application_id)
        except exception.NotFound:
            raise exception.BadRequest('Invalid application')

        routes = self.manager.api.routes().list(
            sysadmin=False, bypass=False, active=True)
        capabilities = self.manager.api.capabilities().list(
            application_id=self.application_id)
        policies = self.manager.api.policies().list(role_id=self.role_id)
        self.capabilities = self._filter_capabilities(routes,
                                                      capabilities,
                                                      policies,
                                                      resources)
        return self.driver.get(id, session) is not None

    def do(self, session, **kwargs) -> None:
        policies = [self._create_policy(self.role_id, capability.id)
                    for capability in self.capabilities]
        session.bulk_save_objects(policies)


class CreateRoles(operation.Operation):

    def pre(self, roles: List[Role], session, **kwargs) -> bool:
        self.roles = roles
        return True

    def do(self, session, **kwargs) -> List[Role]:
        session.bulk_save_objects(self.roles)
        return self.roles


class GetRoleByName(operation.Operation):

    def pre(self, role_name: str, session, **kwargs) -> bool:
        self.roles = self.manager.list(name=role_name)
        return self.roles is not []

    def do(self, session, **kwargs) -> Optional[Role]:
        return self.roles[0]


class Manager(manager.Manager):

    def __init__(self, driver):
        super().__init__(driver)
        self.create = Create(self)
        self.create_policies = CreatePolicies(self)
        self.create_roles = CreateRoles(self)
        self.get_role_by_name = GetRoleByName(self)



from viggocore.common import subsystem
from viggocore.subsystem.constant_for_calculation \
    import resource, router, controller, manager

subsystem = subsystem.Subsystem(resource=resource.ConstantForCalculation,
                                router=router.Router,
                                controller=controller.Controller,
                                manager=manager.Manager)



import json
from typing import Any

from viggocore.common import exception
from viggocore.common.subsystem import entity
from viggocore.database import db
from sqlalchemy import UniqueConstraint


class Application(entity.Entity, db.Model):

    DEFAULT = "default"

    attributes = ['name', 'description', 'settings', 'customized']
    attributes += entity.Entity.attributes

    pagination_column = 'name'

    name = db.Column(db.String(30), nullable=False)
    description = db.Column(db.String(1000), nullable=False)
    _settings = db.Column('settings', db.Text, nullable=False, default='{}',
                          server_default='{}')
    customized = db.Column(db.Boolean(), nullable=False, default=False,
                           server_default='false')

    __table_args__ = (UniqueConstraint('name', name='application_name_uk'),)

    def __init__(self, id, name, description, customized=False,
                 active=True, created_at=None, created_by=None,
                 updated_at=None, updated_by=None, tag=None):
        super().__init__(id, active, created_at, created_by,
                         updated_at, updated_by, tag)
        self.name = name
        self.description = description
        self.customized = customized

    def _has_setting(self, key: str) -> bool:
        return self.settings.get(key) is not None

    def remove_setting(self, key: str):
        if not self._has_setting(key):
            raise exception.BadRequest(f"Erro! Setting {key} not exists")

        settings = self.settings
        value = settings.pop(key)
        self._save_settings(settings)

        return value

    def update_setting(self, key: str, value: Any):
        settings = self.settings
        settings[key] = value
        self._save_settings(settings)
        return value

    @property
    def settings(self):
        try:
            settings_str = '{}' if self._settings is None else self._settings
            return json.loads(settings_str)
        except Exception:
            return {}

    def _save_settings(self, settings: dict):
        self._settings = json.dumps(settings, default=str)



# estilo mais padrão com tracejado entre as linhas da tabela sendo
# pontilhados (usado nos relatórios de estoque de matéria prima e de
# produto por exemplo)
STYLE = """
<style type="text/css">
    .title-table-category  {
        font-size: 10px;
        font-weight: bold;
    }

    .title-table-resume {
        font-size: 10px;
        font-weight: bold;
        text-align: center;
    }

    .div-table {
        display: flex;
        flex-direction: column;
    }

    .table {
        width: 100%;
        height: 100%;
        border-collapse: collapse;
        text-align: center;
    }

    .tr-table-top {
        border-bottom: 0.5px solid gray;
        border-top: 0.5px solid gray;
    }

    .tr-table-center {
        border-top: 0.0px solid gray;
        border-bottom: 0.0px solid gray;
    }

    .tr-table-bottom {
        border-top: 0px solid gray;
    }

    .td1-table {
        text-align: left;
        font-size: 10px;
        padding-left: 30px;
        padding-right: 0px !important;
    }

    .td2-table {
        text-align: right;
        font-size: 10px;
        padding-top: 1px;
        padding-bottom: 1px;
    }

    .tr-table-border {
        border-top: 1px dotted gray;
        border-bottom: 1px dotted gray;
    }
</style>"""  # noqa


from viggocore.common import manager, exception
from viggocore.common.subsystem import operation
from viggocore.common.subsystem.pagination import Pagination
from viggocore.subsystem.capability_functionality.resource \
    import CapabilityFunctionality
from viggocore.subsystem.functionality.resource \
    import Functionality
from sqlalchemy import or_


class GetAvailableFunctionalitys(operation.List):

    def pre(self, **kwargs):
        self.application_id = kwargs.get('application_id', None)
        if self.application_id is None:
            raise exception.BadRequest('Application_id is required')
        return super().pre(**kwargs)

    def do(self, session, **kwargs):
        query = session.query(Functionality). \
            join(CapabilityFunctionality,
                 CapabilityFunctionality.functionality_id == Functionality.id,
                 isouter=True) \
            .filter(
                or_(CapabilityFunctionality.application_id != self.application_id,  # noqa
                    CapabilityFunctionality.application_id == None))  # noqa
        query = self.manager.apply_filters(query, Functionality, **kwargs)
        query = query.distinct()

        dict_compare = {'capability_functionality.': CapabilityFunctionality}
        query = self.manager.apply_filters_includes(
            query, dict_compare, **kwargs)
        query = query.distinct()

        total_rows = None
        if self.manager.with_pagination(**kwargs):
            total_rows = query.count()

        pagination = Pagination.get_pagination(Functionality, **kwargs)
        if pagination.order_by is not None:
            pagination.adjust_order_by(Functionality)
        query = self.driver.apply_pagination(query, pagination)
        result = query.all()

        return (result, total_rows)


class GetSelectedFunctionalitys(operation.List):

    def pre(self, **kwargs):
        self.application_id = kwargs.get('application_id', None)
        if self.application_id is None:
            raise exception.BadRequest('Application_id is required')
        return super().pre(**kwargs)

    def do(self, session, **kwargs):
        query = session.query(Functionality, CapabilityFunctionality.id). \
            join(CapabilityFunctionality,
                 CapabilityFunctionality.functionality_id == Functionality.id,
                 isouter=True) \
            .filter(
                CapabilityFunctionality.application_id == self.application_id)
        query = self.manager.apply_filters(query, Functionality, **kwargs)
        query = query.distinct()

        dict_compare = {'capability_functionality.': CapabilityFunctionality}
        query = self.manager.apply_filters_includes(
            query, dict_compare, **kwargs)
        query = query.distinct()

        total_rows = None
        if self.manager.with_pagination(**kwargs):
            total_rows = query.count()

        pagination = Pagination.get_pagination(Functionality, **kwargs)
        if pagination.order_by is not None:
            pagination.adjust_order_by(Functionality)
        query = self.driver.apply_pagination(query, pagination)
        result = query.all()

        return (result, total_rows)


class Delete(operation.Delete):

    def pre(self, session, id, **kwargs):
        super().pre(session, id=id)
        policies = self.manager.api.policy_functionalities().list(
            capability_functionality_id=id)
        if policies:
            message = 'You can\'t remove this capability because' + \
                ' there are policies associated'
            raise exception.BadRequest(message)
        return True


class Manager(manager.CommonManager):

    def __init__(self, driver):
        super(Manager, self).__init__(driver)
        self.get_available_functionalities = GetAvailableFunctionalitys(self)
        self.get_selected_functionalities = GetSelectedFunctionalitys(self)
        self.delete = Delete(self)



import flask
from viggocore.common import exception, utils


def consulta_cpf_cnpj(self):
    filters = self._filters_parse()

    try:
        response = self.manager.consulta_cpf_cnpj(**filters)
    except exception.ViggoCoreException as exc:
        return flask.Response(response=exc.message, status=exc.status)

    return flask.Response(
        response=utils.to_json(response),
        status=200,
        content_type='application/json',
    )



from viggocore.common.subsystem import manager, entity
from sqlalchemy import func, and_, or_
from datetime import datetime as datetime1
import datetime as datetime2


class CommonManager(manager.Manager):

    def __init__(self, driver):
        super(CommonManager, self).__init__(driver)

    def apply_filters_includes(self, query, dict_compare, **kwargs):
        for k, v in kwargs.items():
            if '.' in k:
                # attribute é o campo da classe do include que deve ser filtrado
                attribute = k.split('.')[-1]
                # individual é a chave que irá identificar qual resource usar
                individual = k.split('.')[0]
                # pega o resource pelo campo "individual" concatenado
                # com o ponto final, pois todos os dict_compare até
                # o momento estão com o ponto final
                resource = dict_compare.get(individual + '.', None)
                # se não encontrar o resource com o ponto ai vai buscar sem,
                # para que os próximos dict_compare possam ser implementados
                # sem o ponto final
                if resource is None:
                    resource = dict_compare.get(individual, None)

                # se resource não for encontrado retorna a query sem filtrar
                if resource is not None and hasattr(resource, attribute):
                    # verifica se não foi passada uma lista de valores
                    if not isinstance(v, list):
                        query = self.apply_filter_individual(
                            query, resource, attribute, v)
                    else:
                        query = self.apply_filter_list(
                            query, resource, attribute, v)

        return query

    def apply_filter_individual(self, query, resource, k, v):
        # verifica se o valor é do tipo string
        isinstance_aux = isinstance(v, str)
        # pega o atributo da classe para verificar se o campo da
        # classe é do tipo string
        attr = getattr(resource, k)

        if self.driver._is_id(k):
            query = query.filter(getattr(resource, k) == v)
        elif k == 'tag':
            # TODO(JorgeSilva): definir o caractere para split
            values = v
            if len(v) > 0 and v[0] == '#':
                values = v[1:]
            values = values.split(',')
            filter_tags = []
            for value in values:
                filter_tags.append(
                    getattr(resource, k)
                    .like('%#' + str(value) + ' %'))
            query = query.filter(or_(*filter_tags))
        # tipo data
        elif isinstance_aux and self.__isdate(v):
            day, next_day = self.__get_day_and_next_day(v)
            query = query.filter(
                and_(
                    or_(getattr(resource, k) < next_day,
                        getattr(resource, k).is_(None)),
                    or_(getattr(resource, k) >= day,
                        getattr(resource, k).is_(None))))
        # tipo string foi passado '%' no valor
        elif isinstance_aux and '%' in v:
            normalize = func.viggocore_normalize
            query = query.filter(
                normalize(getattr(resource, k)).ilike('%' + normalize(v)))
        # tipo string não foi passado '%' no valor
        elif isinstance_aux and attr.type.python_type is str:
            normalize = func.viggocore_normalize
            query = query.filter(
                normalize(getattr(resource, k)) == normalize(v))
        # se for qualquer outro filtro usa esse daqui
        else:
            query = query.filter(getattr(resource, k) == v)

        return query

    def apply_filter_list(self, query, resource, k, v):
        # verifica se o valor é do tipo string
        isinstance_aux = isinstance(v[0], str)
        # pega o atributo da classe para verificar se o campo da
        # classe é do tipo string
        attr = getattr(resource, k)

        # tipo string foi passado '%' no valor
        if self.driver._is_id(k):
            query = query.filter(getattr(resource, k).in_(v))
        elif isinstance_aux and '%' in v[0]:
            normalize = func.viggocore_normalize
            # monta uma lista com os filtros para aplicar o "or"
            filtro = [
                normalize(getattr(resource, k)).ilike('%' + normalize(v_aux))
                for v_aux in v]
            query = query.filter(or_(*filtro))
        # tipo string não foi passado '%' no valor
        elif isinstance_aux and attr.type.python_type is str:
            normalize = func.viggocore_normalize
            # monta uma lista com os filtros para aplicar o "or"
            filtro = [
                normalize(getattr(resource, k)) == normalize(v_aux)
                for v_aux in v]
            query = query.filter(or_(*filtro))
        # se for qualquer outro filtro usa esse daqui
        else:
            query = query.filter(getattr(resource, k).in_(v))

        return query

    def apply_filters(self, query, resource, **kwargs):
        for k, v in kwargs.items():
            if '.' not in k and hasattr(resource, k):
                # verifica se não foi passada uma lista de valores
                if not isinstance(v, list):
                    query = self.apply_filter_individual(
                        query, resource, k, v)
                else:
                    query = self.apply_filter_list(
                        query, resource, k, v)

        return query

    def with_pagination(self, **kwargs):
        require_pagination = kwargs.get('require_pagination', False)
        page = kwargs.get('page', None)
        page_size = kwargs.get('page_size', None)

        if None not in [page, page_size] and require_pagination is True:
            return True
        return False

    def __isdate(self, data, format="%Y-%m-%d"):
        res = True
        try:
            res = bool(datetime1.strptime(data, format))
        except ValueError:
            res = False
        return res

    def __get_day_and_next_day(self, data, format="%Y-%m-%d"):
        day = datetime1.strptime(data, format)
        next_day = day + datetime2.timedelta(days=1)
        return (day, next_day)

    def apply_filter_de_ate(self, resource, query, de, ate):
        inicio = datetime1.strptime(de, entity.DATE_FMT)
        fim = datetime1.strptime(ate, entity.DATE_FMT) +\
            datetime2.timedelta(days=1)
        return query.filter(
            and_(resource.created_at > inicio, resource.created_at < fim))

    # trata os campos "de" e "ate"
    def _convert_de_ate(self, is_date, **kwargs):
        de = kwargs.get('de', None)
        ate = kwargs.get('ate', None)
        inicio = None
        fim = None

        if de and ate:
            try:
                inicio = datetime1.strptime(de.replace(' ', '+'), '%Y-%m-%d%z')
                fim = datetime1.strptime(ate.replace(' ', '+'), '%Y-%m-%d%z') \
                    + datetime2.timedelta(days=1)

                if is_date:
                    de = de[:-5]
                    ate = ate[:-5]
                    inicio = datetime1.strptime(de, entity.DATE_FMT)
                    fim = datetime1.strptime(ate, entity.DATE_FMT) +\
                        datetime2.timedelta(days=1)

            except Exception:
                inicio = datetime1.strptime(de, entity.DATE_FMT)
                fim = datetime1.strptime(ate, entity.DATE_FMT) +\
                    datetime2.timedelta(days=1)

        return (inicio, fim)

    # função criada para filtrar na listagem um campo do tipo Date ou DateTime
    def apply_filter_de_ate_with_timezone(
            self, resource, query, **kwargs):
        attribute = kwargs.get('attribute', 'created_at')
        is_date = str(resource.__getattribute__(resource, attribute).type) == \
            'DATE'
        (de, ate) = self._convert_de_ate(is_date, **kwargs)
        if de and ate:
            if hasattr(resource, attribute):
                query = query.filter(
                    and_(getattr(resource, attribute) >= de,
                         getattr(resource, attribute) < ate))
        return query

    # função criada para filtrar na listagem de uma entidade mais
    # de um campo do tipo Date ou DateTime
    def apply_filter_multiple_de_ate_with_timezone(
            self, resource, query, **kwargs):
        attributes_filter = kwargs.get('attributes_filter', '')
        attributes_filter = attributes_filter.split(',')
        de_filter = kwargs.get('de_filter', '')
        de_filter = de_filter.split(',')
        ate_filter = kwargs.get('ate_filter', '')
        ate_filter = ate_filter.split(',')

        tamanho = len(attributes_filter)

        if not (tamanho == len(de_filter) and tamanho == len(ate_filter)):
            return query

        for i in range(tamanho):
            attribute = attributes_filter[i]
            if attribute == '':
                break
            is_date = str(
                resource.__getattribute__(resource, attribute).type) == 'DATE'
            data = {
                'de': de_filter[i],
                'ate': ate_filter[i]
            }
            (de, ate) = self._convert_de_ate(is_date, **data)
            if de and ate:
                if hasattr(resource, attribute):
                    query = query.filter(
                        and_(getattr(resource, attribute) >= de,
                             getattr(resource, attribute) < ate))
        return query



from viggocore.database import db
from viggocore.common.subsystem import entity


class AlembicVersionHistory(entity.Entity, db.Model):

    attributes = ['titulo', 'anterior_id', 'atual_id',
                  'data_criacao', 'data_execucao', 'tipo']
    attributes += entity.Entity.attributes

    ordem = db.Column(
        db.Integer, nullable=False, primary_key=True, autoincrement=True)
    titulo = db.Column(db.String(200), nullable=False)
    anterior_id = db.Column(db.String(50), nullable=True)
    atual_id = db.Column(db.String(50), nullable=False)
    data_criacao = db.Column(db.DateTime, nullable=False)
    data_execucao = db.Column(db.DateTime, nullable=False)
    tipo = db.Column(db.String(50), nullable=False)

    __tablename__ = 'alembic_version_history'

    def __init__(self, id, titulo, atual_id, data_criacao,
                 data_execucao, tipo, anterior_id=None,
                 active=True, created_at=None, created_by=None,
                 updated_at=None, updated_by=None, tag=None):
        super().__init__(id, active, created_at, created_by,
                         updated_at, updated_by, tag)
        self.titulo = titulo
        self.anterior_id = anterior_id
        self.atual_id = atual_id
        self.data_criacao = data_criacao
        self.data_execucao = data_execucao
        self.tipo = tipo

    @classmethod
    def individual(self):
        return 'alembic_version_history'



import flask

from viggocore.common import exception, utils
from viggocore.common.subsystem import controller


class Controller(controller.Controller):

    def __init__(self, manager, resource_wrap, collection_wrap):
        super(Controller, self).__init__(
            manager, resource_wrap, collection_wrap)

    def get_token_id(self):
        return flask.request.headers.get('token')

    def get_token(self, token_id):
        return self.manager.api.tokens().get(id=token_id)

    def get_available_functionalities(self):
        filters = self._filters_parse()

        try:
            filters = self._parse_list_options(filters)
            functionalities, total_rows = self.manager\
                .get_available_functionalities(**filters)

            collection = self._entities_to_dict(
                functionalities, self._get_include_dicts(filters))

            response = {'capability_functionality': collection}

            page = filters.get('page', None)
            page_size = filters.get('page_size', None)

            if total_rows is not None:
                response.update(
                    {'pagination': {'page': int(page),
                                    'page_size': int(page_size),
                                    'total': total_rows}})
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

        return flask.Response(response=utils.to_json(response),
                              status=200,
                              mimetype="application/json")

    def get_selected_functionalities(self):
        filters = self._filters_parse()

        try:
            filters = self._parse_list_options(filters)
            functionalities, total_rows = self.manager\
                .get_selected_functionalities(**filters)

            includes = self._get_include_dicts(filters)
            collection = []
            if len(functionalities) > 0:
                for functionality in functionalities:
                    functionality_aux = functionality[0].to_dict(includes)
                    functionality_aux['policy_functionality_id'] = (
                        functionality[1])
                    collection.append(functionality_aux)

            response = {'capability_functionality': collection}

            page = filters.get('page', None)
            page_size = filters.get('page_size', None)

            if total_rows is not None:
                response.update(
                    {'pagination': {'page': int(page),
                                    'page_size': int(page_size),
                                    'total': total_rows}})
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

        return flask.Response(response=utils.to_json(response),
                              status=200,
                              mimetype="application/json")

    def create_policy_functionalities(self):
        data = flask.request.get_json()
        policy_functionalities = data.get(
            'policy_functionalities', [])
        successes = []
        errors = []

        for pf in policy_functionalities:
            try:
                pf_entity = self.manager.create(**pf)
                successes.append(pf_entity.to_dict())
            except Exception as exc:
                pf['msg_error'] = exc.message
                errors.append(pf)

        response = {'successes': successes, 'errors': errors}

        return flask.Response(response=utils.to_json(response),
                              status=200,
                              mimetype="application/json")



from typing import Dict, List

from viggocore.common.subsystem import Subsystem
from viggocore.common.subsystem.apihandler import Api
from viggocore.common.input import RouteResource
from viggocore.subsystem.role.resource import Role
from viggocore.bootstrap.roles import BootstrapRoles
from viggocore.bootstrap.default.application import BootstrapApplication
from viggocore.bootstrap.default.domain import BootstrapDomain
from viggocore.bootstrap.default.user import BootstrapUser
from viggocore.bootstrap.default.policies import BootstrapPolicies


class BootstrapDefault(object):

    def __init__(self, api: Api) -> None:
        self.bootstrap_roles = BootstrapRoles(api)
        self.bootstrap_application = BootstrapApplication(api)
        self.bootstrap_domain = BootstrapDomain(api)
        self.bootstrap_user = BootstrapUser(api)
        self.bootstrap_policies = BootstrapPolicies(api)

    def execute(self, user_resources: List[RouteResource],
                sysadmin_resources: List[RouteResource],
                sysadmin_exclusive_resources: List[RouteResource]):
        roles = self.bootstrap_roles.execute()
        role_sysadmin = self._get_role_sysadmin(roles)

        application = self.bootstrap_application.execute()
        domain = self.bootstrap_domain.execute(application.id)
        self.bootstrap_user.execute(domain.id, role_sysadmin.id)
        self.bootstrap_policies.execute(application.id,
                                        role_sysadmin.id,
                                        user_resources,
                                        sysadmin_resources,
                                        sysadmin_exclusive_resources)

    def _get_role_sysadmin(self, roles: List[Role]) -> Role:
        role = next((role for role in roles if role.name == Role.SYSADMIN),
                    None)
        if not role:
            raise Exception()
        return role



from setuptools import setup, find_packages
from viggocore._version import version


REQUIRED_PACKAGES = [
    'alembic==1.6.2',
    'amqp==5.0.6',
    'APScheduler==3.7.0',
    'attrs==21.2.0',
    'billiard==3.6.4.0',
    'celery==5.0.5',
    'certifi==2021.10.8',
    'chardet==4.0.0',
    'click==7.1.2',
    'click-didyoumean==0.0.3',
    'click-plugins==1.1.1',
    'click-repl==0.1.6',
    'colorama==0.4.4',
    'decorator==5.0.8',
    'Flask==2.0.0',
    'Flask-Migrate==2.7.0',
    'Flask-RBAC==0.5.0',
    'Flask-SQLAlchemy==2.5.1',
    'gabbi==2.2.0',
    'greenlet==1.1.0',
    'idna==2.10',
    'iniconfig==1.1.1',
    'itsdangerous==2.0.0',
    'Jinja2==3.0.0',
    'jsonpath-rw==1.4.0',
    'jsonpath-rw-ext==1.2.2',
    'kombu==5.0.2',
    'Mako==1.1.4',
    'MarkupSafe==2.0.0',
    'packaging==20.9',
    'pbr==5.6.0',
    'pika==1.2.0',
    'Pillow==9.1.0',
    'pluggy==0.13.1',
    'ply==3.11',
    'prompt-toolkit==3.0.18',
    'py==1.10.0',
    'pyparsing==2.4.7',
    'pytest==6.2.4',
    'python-dateutil==2.8.1',
    'python-editor==1.0.4',
    'pytz==2021.1',
    'PyYAML==6.0',
    'requests==2.25.1',
    'six==1.16.0',
    'sparkpost==1.3.9',
    'SQLAlchemy==1.4.15',
    'toml==0.10.2',
    'tzlocal==2.1',
    'urllib3==1.26.5',
    'vine==5.0.0',
    'wcwidth==0.2.5',
    'Werkzeug==2.0.0',
    'wsgi-intercept==1.9.2',
]

setup(
    name='viggocore',
    version=version,
    summary='Viggocore Framework',
    url='https://github.com/viggosistemas/viggocore-backend',
    author='Samuel de Medeiros Queiroz, Francois Oliveira, ' + \
        'Jorge Júnio da Silva',
    author_email='samueldmq@gmail.com, oliveira.francois@gmail.com, ' + \
        'jorgejunioufrn@gmail.com',
    license='Apache-2',
    packages=find_packages(exclude=["tests"]),
    install_requires=REQUIRED_PACKAGES
)



from viggocore.common.subsystem import operation


class Create(operation.Create):

    def pre(self, session, **kwargs):
        validar_user = kwargs.get('validar_user', False)
        if validar_user is True:
            self.manager.valida_user(**kwargs)
        return super().pre(session, **kwargs)



from sqlalchemy import orm, UniqueConstraint

from viggocore.database import db
from viggocore.common.subsystem import entity
from viggocore.common.utils import random_uuid
from viggocore.common import exception


class Functionality(entity.Entity, db.Model):

    CODE_SEQUENCE = 'functionality_seq'

    attributes = ['code', 'name', 'description']
    attributes += entity.Entity.attributes

    code = db.Column(db.Numeric(10), nullable=False, unique=True)
    name = db.Column(db.String(80), nullable=False, unique=True)
    description = db.Column(db.String(1000), nullable=True)

    routes = orm.relationship(
        'FunctionalityRoute',
        backref=orm.backref('functionality_routes'),
        cascade='delete,delete-orphan,save-update')

    def __init__(self, id, code, name, description=None,
                 active=True, created_at=None, created_by=None,
                 updated_at=None, updated_by=None, tag=None):
        super().__init__(id, active, created_at, created_by,
                         updated_at, updated_by, tag)
        self.code = code
        self.name = name
        self.description = description

    def _get_route(self, route_id):
        route = next((r for r in self.routes
                      if r.route_id == route_id), None)
        if not route:
            raise exception.BadRequest('ERRO! Route not found!')

        return route

    def add_route(self, route_id):
        try:
            route = self._get_route(route_id)
        except Exception:
            route = FunctionalityRoute(random_uuid(), self.id, route_id)
            self.routes.append(route)

    def rm_route(self, route_id):
        route = self._get_route(route_id)
        self.routes.remove(route)

    def add_routes(self, route_ids):
        for route_id in route_ids:
            self.add_route(route_id)
        return self

    def rm_routes(self, route_ids):
        for route_id in route_ids:
            self.rm_route(route_id)
        return self

    @classmethod
    def individual(cls):
        return 'functionality'

    @classmethod
    def collection(cls):
        return 'functionalities'


class FunctionalityRoute(entity.Entity, db.Model):

    attributes = ['id', 'functionality_id', 'route_id']

    functionality_id = db.Column(
        db.CHAR(32), db.ForeignKey('functionality.id'), nullable=False)
    functionality = orm.relationship(
        'Functionality',
        backref=orm.backref('functionality_route_functionality'),
        viewonly=True)
    route_id = db.Column(
        db.CHAR(32), db.ForeignKey("route.id"), nullable=False)
    route = orm.relationship(
        'Route',
        backref=orm.backref('functionality_route_route'))

    __table_args__ = (UniqueConstraint(
        'functionality_id', 'route_id',
        name=('functionality_route_functionality_id_route_id_uk')),)

    def __init__(self, id, functionality_id, route_id,
                 active=True, created_at=None, created_by=None,
                 updated_at=None, updated_by=None, tag=None):
        super().__init__(id, active, created_at, created_by,
                         updated_at, updated_by, tag)
        self.functionality_id = functionality_id
        self.route_id = route_id

    @classmethod
    def individual(cls):
        return ['functionality_route']

    @classmethod
    def collection(cls):
        return ['functionality_routes']



# estilo mais padrão com tracejado entre as linhas da tabela sendo
# pontilhados (usado nos relatórios de estoque de matéria prima e de
# produto por exemplo)
STYLE = """
<style>
    body {
        font-family: Georgia, 'Times New Roman', Times, serif, sans-serif;
        margin: 15px;
        font-size: 10px;
    }

    .cabecalho {
        display: flex;
        justify-content: space-between;
        align-items: flex-start;
        padding-bottom: 5px;
        margin-bottom: 15px;
    }

    .logo img {
        width: 70px;
        height: auto;
    }

    .centro {
        text-align: center;
        flex-grow: 1;
    }

    .centro h2 {
        margin: 0;
        font-size: 13px;
    }

    .centro p {
        margin: 1px 0;
        font-size: 9px;
    }

    .centro .titulo-relatorio {
        margin-top: 12px;
        font-size: 13px;
        font-weight: bold;
        text-transform: uppercase;
    }

    .emissao {
        font-size: 9px;
        text-align: right;
    }

    table {
        width: 100%;
        border-collapse: collapse;
    }

    th,
    td {
        border: 1px solid #ccc;
        padding: 4px;
        vertical-align: top;
        font-size: 9px;
    }

    th {
        background-color: #f9f9f9;
        color: #000000;
        /* ← Cor preta aplicada aqui */
        text-align: left;
    }

    td.foto {
        text-align: center;
    }

    td.foto img {
        width: 60px;
        height: 60px;
        border-radius: 50%;
        object-fit: cover;
    }

    .motivo {
        white-space: pre-wrap;
        padding-top: 2px;
        font-style: italic;
    }

    .linha-motivo td {
        border-top: none;
        background-color: #f9f9f9;
        font-size: 9px;
    }
</style>"""  # noqa



from datetime import datetime
from enum import Enum
from openpyxl import Workbook, styles, utils
from openpyxl.worksheet.worksheet import Worksheet
from viggocore.common import utils as core_utils
from viggocore.common.subsystem import operation


class MakeExportarXsls(operation.Operation):
    def _map_column(self, column):
        aux = {
            'num': self.columns.index(column) + 1,
            'name': column.get('header_name', ''),
            'width': column.get('width', 50)
        }
        if column.get('total', False) is True:
            aux['footer'] = '=sum()'

        return aux

    def _def_header_style(self, ws, cols, style_name, row_ini=1):
        headers = [col['name'] for col in cols]
        ws.append(headers)

        header_style = styles.NamedStyle(name=style_name)
        header_style.font = styles.Font(bold=True)
        header_style.border = styles.Border(
            top=styles.Side(border_style='thin'),
            right=styles.Side(border_style='thin'),
            bottom=styles.Side(border_style='thin'),
            left=styles.Side(border_style='thin'))
        header_style.alignment = styles.Alignment(horizontal='left',
                                                  vertical='center')

        for col_num in range(len(headers)):
            col = utils.cell.get_column_letter(col_num + 1)
            ws.column_dimensions[col].width = cols[col_num]['width']

        last_col = utils.cell.get_column_letter(len(headers)) + '1'
        cell_range = ws['A1':last_col]

        for cell_rows in cell_range:
            for cell in cell_rows:
                cell.style = header_style

    def format_data_hora(self, data_hora):
        dia = data_hora.day if data_hora.day > 9 else '0' + str(data_hora.day)
        mes = (data_hora.month if data_hora.month > 9
               else '0' + str(data_hora.month))
        return '{}/{}/{}'.format(dia, mes, data_hora.year)

    def _treat_data_type(self, data, col):
        data_type = col.get('type', 'STRING')
        result = ''

        if data_type == 'STRING':
            result = data if data is not None else '-'
            if data is not None:
                result = data.name if isinstance(data, Enum) else data
        elif data_type == 'BOOLEAN':
            result = 'Sim' if data is True else 'Não'
        elif data_type == 'DATE':
            result = datetime.strftime(data, '%d/%m/%Y') \
                if data is not None \
                else '-'
        elif data_type == 'DATETIME':
            result = datetime.strftime(data, '%d/%m/%Y %H:%M') \
                if data is not None \
                else '-'
        elif data_type == 'CURRENCY' or data_type == 'FLOAT':
            decimals = col.get('decimals', None)
            if decimals is not None:
                result = core_utils.to_decimal_n(data, decimals) \
                    if data is not None \
                    else core_utils.to_decimal_n('0', decimals)
            else:
                result = core_utils.to_decimal(data) \
                    if data is not None \
                    else core_utils.to_decimal('0')
        elif data_type == 'INT':
            result = int(data) if data is not None else 0

        return result

    def _map_row(self, row):
        result = []
        for col in self.columns:
            aux = ''
            fields = col.get('field', None)
            if fields is not None:
                fields = fields.split('.')
                if len(fields) > 1:
                    aux = None if row is None else row
                    if row is not None:
                        for field in fields:
                            if aux is not None:
                                aux = aux.__getattribute__(field)
                else:
                    aux = row.__getattribute__(fields[0])

                aux = self._treat_data_type(aux, col)
            result.append(aux)

        return result

    def _make_relatorio(self):
        relatorio = []
        for e in self.entities:
            relatorio.append(self._map_row(e))

        return relatorio

    def _formatar_footer(self, col, cell, lin_ini, qtd_linhas,
                         is_to_bold=True):
        if is_to_bold:
            cell.font = styles.Font(bold=True)

        if ('footer' in col):
            if not col['footer'].startswith('='):
                cell.value = col['footer']
            else:
                formula = col['footer']
                coluna = utils.cell.get_column_letter(
                    col['num'])
                intervalo_celulas = '(' + \
                    str(coluna) + str(lin_ini) + ':' + \
                    str(coluna) + str(qtd_linhas + 1) + ')'
                formula = formula.replace(
                    '()', intervalo_celulas)
                cell.value = formula

    def _adiciona_total_linhas(self, ws, row_num, qtd_linhas):
        # adiciona linha com os totais
        cell = ws.cell(row_num, 1)
        cell.border = styles.Border(
            top=styles.Side(border_style='thin'),
            right=styles.Side(border_style='thin'),
            bottom=styles.Side(border_style='thin'),
            left=styles.Side(border_style='thin'))
        cell.font = styles.Font(bold=True)
        cell.value = f'Total de itens: {qtd_linhas}'

    def _formatar_celulas(self, ws, dados, cols, is_to_bold=True):
        qtd_linhas = len(dados)

        if qtd_linhas > 0:
            lin_ini = 2
            lin_fim = lin_ini + qtd_linhas  # A última linha é de total

            for lin in range(lin_ini, lin_fim + 1):
                # Accounting format
                for col in cols:
                    cell = ws.cell(lin, col['num'])
                    cell.border = styles.Border(
                        top=styles.Side(border_style='thin'),
                        right=styles.Side(border_style='thin'),
                        bottom=styles.Side(border_style='thin'),
                        left=styles.Side(border_style='thin'))

                    if 'fmt' in col:
                        cell.number_format = col['fmt']

                    if 'align' in col:
                        cell.alignment = styles.Alignment(
                            horizontal=col['align'])

                    if 'function' in col:
                        formula = col['function'] \
                            .replace('[row]', str(cell.row))
                        cell.value = formula

                    if (lin == lin_fim):
                        self._formatar_footer(
                            col, cell, lin_ini, qtd_linhas)
        # adiciona quantidade de linhas
        self._adiciona_total_linhas(ws, lin_fim + 1, qtd_linhas)
        return

    def _style_header(self, ws: Worksheet, cols: list) -> None:
        for i in range(1, len(cols) + 1):
            cell = ws.cell(column=i, row=1)
            cell.fill = styles.PatternFill('solid', fgColor='DEEBF7')
            cell.border = styles.Border(
                top=styles.Side(border_style='medium', color='000000'),
                right=styles.Side(border_style='thin', color='000000'),
                bottom=styles.Side(border_style='medium', color='000000'),
                left=styles.Side(border_style='thin', color='000000'))
            cell.alignment = styles.Alignment(horizontal='center')

    def _align_horizontal_center_column(
            self, ws: Worksheet, linhas: list, coluna: int) -> None:
        for i in range(1, len(linhas) + 2):
            cell = ws.cell(column=coluna, row=i)
            cell.alignment = styles.Alignment(horizontal='center')

    def _gen_work_sheet(self, ws: Worksheet):
        # num_fmt = u'_(* #,##0.00_);[Red]_(* (#,##0.00);_(* -_0_0_);_(@'
        cols = list(map(lambda x: self._map_column(x), self.columns))
        self._def_header_style(ws, cols, self.title)

        relatorio = self._make_relatorio()
        for linha in relatorio:
            ws.append(linha)

        self._formatar_celulas(ws, relatorio, cols, False)
        self._style_header(ws, cols)
        self._align_horizontal_center_column(ws, relatorio, 2)

        return True

    def pre(self, session, entities, columns, title, file_title):
        self.entities = entities
        self.columns = columns
        self.title = title

        self.now = datetime.now()
        self.file_name = f'{file_title}-{self.now}.xlsx'
        self.download_folder = \
            self.manager.get_download_folder(entities[0].domain_id)

        return True

    def do(self, session, entities, columns, title, file_title):
        wb = Workbook()
        ws = wb.active
        ws.title = self.title
        self._gen_work_sheet(ws)

        wb.save(filename=self.download_folder + '/' + self.file_name)

        return self.download_folder, self.file_name
